import logging
from sittercommon.machinedata import MachineData

logger = logging.getLogger(__name__)


class HasMachineSitter(object):
    """
    Everything is asynchronous -- always returns a request
    object that can be run later.
    """
    def __init__(self):
        self.hostname = None
        self.datamanager = None
        self.historic_data = []
        self.loaded = False

    def _api_start_task(self, job):
        val = self.datamanager.add_task(job.task_configuration)
        logger.info("Add task %s result: %s" % (job.name, val))
        self.datamanager.reload()
        if job.name in self.datamanager.tasks:
            val = self.datamanager.start_task(
                self.datamanager.tasks[job.name])
            logger.info("Start task %s result: %s" % (job.name, val))
            self.datamanager.reload()
            return val
        else:
            return False

    def _api_stop_task(self, job):
        if job.name in self.datamanager.tasks:
            val = self.datamanager.stop_task(
                self.datamanager.tasks[job.name])
            logger.info("Stop task %s result: %s" % (job.name, val))
            self.datamanager.reload()
            return val
        else:
            return False

    def _api_identify_sitter(self):
        logger.info("Attempting to find a machinesitter at %s" % (
                self.hostname))
        if not self.datamanager:
            self.datamanager = MachineData(self.hostname, 40000)
        else:
            self.datamanager._find_portnum()

        if self.datamanager.url:
            logger.info("Found sitter at %s" % self.datamanager.url)
        else:
            logger.warn("Couldn't find a sitter for %s" % self.hostname)

        return self.datamanager.portnum

    def _api_get_endpoint(self, path):
        return "%s/%s" % (self.datamanager.url,
                                    path)

    def _api_get_stats(self):
        logger.debug("Get stats for %s" % str(self))
        tasks = None
        try:
            tasks = self.datamanager.reload()
        except:
            self.loaded = False
            import traceback
            traceback.print_exc()
            logger.error(traceback.format_exc())

        if tasks != None:
            self.loaded = True
        else:
            self.loaded = False

        logger.debug("Get stats for %s result: %s" % (str(self),
                                                     self.loaded))
        return self.loaded

    def _get_machinename(self):
        if self.datamanager:
            return "%s:%s" % (self.datamanager.hostname,
                              self.datamanager.portnum)
        else:
            return self.hostname


class MonitoredMachine(HasMachineSitter):
    """
    An interface for a single
    machine to monitor.  Some functions
    Should be implemented per cloud provider.
    Note: It it assumed that the MachineMonitor
    keeps all MonitoredMachines up to date, and that
    with the exception of functions explicitly about
    downloading data, all calls are accessing LOCAL
    CACHED data and NOT making network calls.
    """
    def __init__(self, config, machine_number=0,
                  *args, **kwargs):
        super(MonitoredMachine, self).__init__(*args, **kwargs)
        self.config = config
        self.running_tasks = []
        self.machine_number = machine_number
        self.hostname = self.config.hostname
        self.detected_sitter_failures = 0

        # Used for deployment purposes by a JobFiller
        self.state = None

    def is_in_deployment(self):
        if not self.state:
            return False

        if self.state.get_state() == 6:
            return False

        return True

    def get_tasks(self):
        return self.datamanager.tasks

    def get_running_tasks(self):
        """
        Return cached data about running task status
        """
        if not self.datamanager:
            return []

        return [task for task in self.datamanager.tasks.values() if \
                    task["running"]]

    def start_task(self, job):
        if self.is_initialized():
            logger.info("Starting a task %s on %s" % (job.name, str(self)))
            self._api_start_task(job)

    def stop_task(self, job):
        if self.is_initialized():
            logger.info("Stopping a task %s on %s" % (job.name, str(self)))
            self._api_stop_task(job)

    def initialize(self):
        return self._api_identify_sitter()

    def is_initialized(self):
        if not self.datamanager:
            return False

        return self.datamanager.url != ""

    def has_loaded_data(self):
        return self.is_initialized() and self.loaded

    def __str__(self):
        return self._get_machinename()

    def __repr__(self):
        return str(self)



#!/usr/bin/python -u
"""
Cerebro client tool
"""
import json
import os
import sys

from sittercommon import arg_parser
from sittercommon.utils import (
    update_job, job_cfg_update, change_debug_level,
    update_idle_limit)

COMMANDS = [update_job,
            job_cfg_update,
            change_debug_level,
            update_idle_limit]


def get_parser(default_options):
    parser = arg_parser.ArgumentParser(
        description="Cerebro Command Line Interaction Tool")

    parser.add_argument(
        "--clustersitter-url", dest="clustersitter_url",
        help="URL to the root of the clustersitter",
        default=default_options.get('clustersitter_url'))

    return parser


def main(sys_args=None):
    if not sys_args:
        sys_args = sys.argv[1:]

    default_options = {}
    CFG_FILE = os.path.expanduser("~/.cerebro.cfg")

    try:
        default_data = open(CFG_FILE).read()
        default_options = json.loads(default_data)
    except:
        pass

    command_parsers = {}
    print default_options
    parser = get_parser(default_options)
    subparsers = parser.add_subparsers(help='Cerebro Command')
    for command in COMMANDS:
        command_parsers[command.get_command()] = subparsers.add_parser(
            command.get_command(),
            help=command.get_help_string())
        command.get_parser(command_parsers[command.get_command()])
        command_parsers[command.get_command()].set_defaults(
            command=command.get_command())

    args = parser.parse_args(sys_args)

    if not args.command:
        print "No Command Given"
        print parser.print_help()
        sys.exit(0)

    if not args.clustersitter_url:
        print "Please pass a clustersitter URL the first time you run cerebro"
        sys.exit(0)

    default_options['clustersitter_url'] = args.clustersitter_url
    default_file = open(CFG_FILE, 'w')
    default_file.write(json.dumps(default_options))
    default_file.close()

    for command in COMMANDS:
        if command.get_command() == args.command:
            newargs = dict(args._get_kwargs())
            del newargs['command']
            command.run_command(**newargs)



import logging
import socket
import threading
import time
from datetime import datetime, timedelta

from deploymentrecipe import MachineSitterRecipe
from monitoredmachine import MonitoredMachine
from eventmanager import ClusterEventManager

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class StateMachine(object):
    states = {}

    def __init__(self):
        self.state = 0
        self.last_state = 0

    def get_state(self):
        return self.state

    def set_error(self):
        self.set_state(-1)

    def set_state(self, new_state):
        self.last_state = self.state
        self.state = new_state

    def next(self):
        self.set_state(self.state + 1)

    def __str__(self):
        return "%s: %s" % (self.state, self.states[self.state])


class MachineDeploymentState(StateMachine):
    states = {
        -1: 'Error',
         0: 'None',
         1: 'Provider.Launching',
         2: 'TaskSitter.Deployment',
         3: 'Job.Deployment',
         4: 'Job.Launching',
         5: 'AddingMonitoring',
         6: 'Ready'
        }


class JobFillerState(StateMachine):
    states = {
        -1: 'Error',
         0: 'CreatingResources',
         1: 'EnsureDNS',
         2: 'DeployingMonitoringCode',
         3: 'DeployingJobCode',
         4: 'LaunchingTasks',
         5: 'AddingtoMonitoring',
         6: 'EnsureBaseDNS',
         7: 'Reboot Dependent Jobs',
         8: 'Done'
         }


class JobFiller(object):
    def __init__(self, num_cores, job, zone, idle_machines=None,
                 raw_machines=None, reboot_task=False, post_callback=None,
                 fail_on_error=False):
        self.num_cores = num_cores
        self.job = job
        self.zone = zone
        self.machines = idle_machines or []
        self.state = JobFillerState()
        self.thread = None
        self.end_time = None
        self.reboot_task = reboot_task
        self.post_callback = post_callback
        self.fail_on_error = fail_on_error

        if not raw_machines:
            raw_machines = []

        self.machine_states = {}

        for machine in self.machines:
            self.machine_states[machine] = MachineDeploymentState()
            self.machine_states[machine].set_state(3)

        for machine in raw_machines:
            self.machine_states[machine] = MachineDeploymentState()
            self.machine_states[machine].set_state(1)
            self.machines.append(machine)

    def __str__(self):
        return "%s cores in %s for %s" % (self.num_cores,
                                          self.zone, self.job)

    def num_remaining(self):
        if self.is_done():
            return 0
        else:
            return self.num_cores

    def start(self):
        self.thread = threading.Thread(
            target=self.run,
            name="JobFiller-%s:%s:%s" % (
                self.job.name,
                self.zone,
                self.num_cores))

        self.thread.start()

    def is_done(self):
        return self.state.get_state() == 8

    def run(self):
        logger.info("Starting JobFiller")
        release_attempts = 1
        while self.state.get_state() != 8:
            state = self.state.get_state()
            logger.info("Running State: %s, attempt #%s" % (
                    str(self.state),
                    release_attempts))

            try:
                if state == 0:
                    self.run_create_resources()
                elif state == 1:
                    self.ensure_dns()
                elif state == 2:
                    self.deploy_monitoring_code()
                elif state == 3:
                    self.deploy_job_code()
                elif state == 4:
                    self.launch_tasks()
                elif state == 5:
                    self.add_to_monitoring()
                elif state == 6:
                    self.ensure_dns(do_basename=True)
                elif state == 7:
                    self.reboot_dependent_jobs()
            except:
                release_attempts += 1
                import traceback
                traceback.print_exc()
                logger.error(traceback.format_exc())

                if release_attempts > 10 or self.fail_on_error:
                    logger.info("Job Filler: Failed")
                    ClusterEventManager.handle(
                        "Failed Filling: %s" % str(self))

                    if self.post_callback:
                        self.post_callback(success=False)

                    return False

        ClusterEventManager.handle(
            "Completed Filling: %s" % str(self))
        logger.info("Job Filler: Done!")
        self.end_time = datetime.now()

        if self.post_callback:
            self.post_callback(success=True)

        return True

    def run_create_resources(self):
        if len(self.machines) < self.num_cores:
            #!MACHINEASSUMPTION! Should calculate machines/cores
            # (input_machines.num_cores).
            new_machine_count = self.num_cores - len(self.machines)
            # Then spin some up!
            if self.launch_machines(new_machine_count):
                self.state.next()
            else:
                logger.warn(
                    "Couldn't launch machines for %s?" % self.job.name)
        else:
            self.state.next()

    def ensure_dns(self, do_basename=False):
        basename = "%s.%s" % (self.zone, self.job.dns_basename)
        provider = self.job.sitter.dns_provider
        if not self.job.dns_basename or not provider:
            self.state.next()
            return

        num_machines_total = self.job.get_num_required_machines_in_zone(
            self.zone, self.state)

        try:
            records = provider.get_records(hostName=basename)
        except:
            import traceback
            logger.error("Couldn't download records for %s" % basename)
            logger.error(traceback.format_exc())
            records = []

        # Find out what records exist underneath this basename
        # and ensure the machine objects have them
        record_by_ip = {}
        used_prefixes = []
        records_for_basename = []
        for record in records:
            if record['record'] == basename:
                records_for_basename.append(record['value'])

            pieces = record['record'].split('.')
            if pieces[0].isdigit() and '.'.join(pieces[1:]) == basename:
                used_prefixes.append(int(pieces[0]))
                record_by_ip[record['value']] = record['record']

        for machine in self.machines:
            ip = socket.gethostbyname(machine.config.hostname)
            machine.config.ip = ip
            # Ensure 2 things:
            # 1) x.basename exists for this machine
            # 2) basename includes this machine's IP as an a record

            # Part 1
            if ip in record_by_ip:
                machine.config.dns_name = record_by_ip[ip]
            else:
                # We need a new name!  Get the first valid one
                new_prefix = None

                for i in range(num_machines_total):
                    if i not in used_prefixes:
                        new_prefix = i
                        used_prefixes.append(i)
                        break

                if new_prefix == None:
                    # This only happens if the state of DNS
                    # records somehow doesn't match the state
                    # of this job, which shouldn't happen
                    if used_prefixes:
                        new_prefix = max(used_prefixes) + 1
                    else:
                        new_prefix = 0

                machine.config.dns_name = "%s.%s" % (new_prefix,
                                                     basename)
                logger.info("Assigning %s to %s" % (
                        machine.config.dns_name, ip))

                if not provider.add_record(ip,
                                          machine.config.dns_name):
                    logger.error("Couldn't assign DNS for %s" % (
                            machine.config.dns_name))

            # Part 2
            if do_basename:
                """
                We should only add to the basename once the machine
                is up and running.  Otherwise we're effectively adding
                this machine 'to the pool' before its ready, potentially
                causing problems.
                """
                if ip not in records_for_basename:
                    logger.info("Adding %s to %s" % (ip, basename))
                    if not provider.add_record(ip, basename):
                        logger.error("Couldn't assign DNS for %s -> %s" % (
                                ip,
                                basename))

        self.state.next()

    def deploy_monitoring_code(self):
        # TODO Parallelize this somehow
        for machine in self.machines:
            recipe = self.job.sitter.build_recipe(
                MachineSitterRecipe,
                machine,
                post_callback=None,
                options={'log_location': self.job.sitter.log_location})

            self._do_recipe_deployment(2, 3, machine,
                                       recipe)
        self.state.next()

    def _do_recipe_deployment(self, old_state,
                              new_state,
                              machine,
                              recipe):
        # We'll try and deploy to a machine
        # for a total of 2 minutes.  This is random
        # and probably a terrible way to limit
        # how many times we try and deploy, but
        # I can't think of something better right now.  *shrug*.
        start_time = datetime.now()

        while self.machine_states[machine].get_state() <= old_state:
            self.machine_states[machine].set_state(old_state)
            recipe.connect()
            val = recipe.deploy()
            if val:
                self.machine_states[machine].set_state(new_state)
            else:
                logger.warn(
                    "Couldn't deploy code to %s?" % (
                        str(machine)))

                if datetime.now() - start_time > timedelta(minutes=2):
                    self.machine_states[machine].set_state(new_state)
                    logger.error("Giving up deploying to %s" % str(
                            machine))

                    raise Exception("Failed to deploy!")

                time.sleep(1)

    def deploy_job_code(self):
        # TODO Parallelize this somehow
        for machine in self.machines:
            while self.machine_states[machine].get_state() <= 3:
                self.machine_states[machine].set_state(3)

                if not self.job.deployment_recipe:
                    self.machine_states[machine].set_state(4)
                    continue

                recipe = self.job.sitter.build_recipe(
                    self.job.deployment_recipe,
                    machine,
                    post_callback=None,
                    options=self.job.recipe_options,
                    given_logger=logger)

                self._do_recipe_deployment(3, 4,
                                           machine,
                                           recipe)

        self.state.next()

    def reboot_dependent_jobs(self):
        if not self.reboot_task:
            self.state.next()
            return

        jobs = self.job.find_dependent_jobs()
        for machine in self.machines:
            task_names = [task['name'] for task in machine.get_running_tasks()]
            for job in jobs:
                if job.name in task_names:
                    logger.info("Rebooting %s" % job.name)
                    machine.stop_task(job)
                    machine.start_task(job)

        self.state.next()

    def launch_tasks(self):
        # TODO Parallelize this somehow
        for machine in self.machines:
            while self.machine_states[machine].get_state() <= 4:
                self.machine_states[machine].set_state(4)
                machine.initialize()

                tasks = machine.get_running_tasks()
                task_names = [task['name'] for task in tasks]
                already_running = self.job.name in task_names

                if self.reboot_task or already_running:
                    machine.stop_task(self.job)

                logger.debug("Starting %s" % self.job.name)
                val = machine.start_task(self.job)

                if not val:
                    # Couldn't start the task, which is odd
                    # Should bail out?
                    break

                # Now verify that its started
                machine._api_get_stats()
                tasks = machine.get_running_tasks()
                task_names = [task['name'] for task in tasks]
                if self.job.name in task_names:
                    self.machine_states[machine].set_state(5)
                else:
                    logger.warn(
                        "Tried to start %s on %s but failed?" % (
                            self.job.name, str(machine)))

        self.state.next()

    def add_to_monitoring(self):
        # Ensure the machines aren't already monitored
        machines_to_add = []
        # TODO move this to a function in clustersitter.py
        # def check_already_monitored or some such
        for machine in self.machines:
            found = False
            for monitor, _ in self.job.sitter.state.monitors:
                if machine in monitor.monitored_machines:
                    found = True
                    break

            if not found:
                machines_to_add.append(machine)

        if machines_to_add:
            self.job.sitter.add_machines(machines_to_add, update_dns=False)

        # Now wait for the machines to actually be monitored
        for machine in self.machines:
            ready = False
            while not ready:
                loaded_data = machine.has_loaded_data()
                has_task = self.job.name in machine.get_tasks()
                if self.job.name == "Machine Doctor Redeployer":
                    has_task = True

                ready = loaded_data and has_task
                if not ready:
                    logger.debug(
                        "Waiting for machine to be actually monitored...")
                    time.sleep(0.1)
                else:
                    logger.debug("%s has our task, good to go!" % machine)

        self.state.next()

    def launch_machines(self, new_machine_count):
        provider = self.job.sitter.state.provider_by_zone[self.zone]
        mem_per_job = self.job.deployment_layout[self.zone]['mem']

        machineconfigs = provider.fill_request(zone=self.zone,
                                               cpus=new_machine_count,
                                               mem_per_job=mem_per_job)
        if not machineconfigs:
            return False

        machines = [MonitoredMachine(m) for m in machineconfigs]

        for machine in machines:
            self.machine_states[machine] = MachineDeploymentState()

        self.machines.extend(machines)

        return True



#!/usr/bin/python -u
"""
This is the main file of the machinesitter application.

It's job is to manage multiple tasks on a machine -- starting
and stopping them at the whim of an administrator.
"""

import sittercommon.arg_parser as argparse
import machinemanager


import os
import simplejson
import sys
import time


def parse_args(args):
    """ Parse command line args."""

    parser = argparse.ArgumentParser(
        description="Start the machine sitter daemon")

    parser.add_argument('--task_definitions', dest='taskfile',
                        help='The location of the task definition file')

    parser.add_argument('--log-location', dest='log_location',
                        help='Where logs should go',
                        default='/tmp')

    parser.add_argument("--daemon", dest="daemon",
                        default=False,
                        action="store_true",
                        help='Daemonize and split from launching shell')

    return parser.parse_args(args=args)


def daemonize():
    pid = os.fork()
    if pid > 0:
        sys.exit(0)

    os.setsid()
    os.chdir("/")

    pid = os.fork()
    if pid > 0:
        sys.exit(0)

    os.umask(0)
    sys.stdout.flush()
    sys.stderr.flush()
    sys.stdin.close()

    print "Sitter PID: %s" % os.getpid()


def main(sys_args=None):

    if not sys_args:
        sys_args = sys.argv[1:]

    orig_dir = os.getcwd()

    args = parse_args(sys_args)

    if args.daemon:
        print "Daemonizing"
        daemonize()

    if args.taskfile:
        config = simplejson.load(open(args.taskfile))
    else:
        config = {'task_definitions': {}}

    if not 'log_location' in config:
        config['log_location'] = args.log_location

    try:
        os.makedirs(config['log_location'])
    except:
        pass

    manager = machinemanager.MachineManager(args.taskfile,
                                            config['log_location'],
                                            machine_sitter_starting_port=40000,
                                            task_sitter_starting_port=50000,
                                            launch_location=orig_dir,
                                            daemon=args.daemon)

    task_definitions = config['task_definitions']

    for task in task_definitions:
        manager.add_new_task(task)

    manager.start()

    # wait forever
    while True:
        time.sleep(1)



import logging
import re
import requests
import simplejson
import socket
import time
import urllib

logger = logging.getLogger(__name__)


class MachineData(object):
    def __init__(self, hostname, starting_port):
        self.hostname = hostname
        self.portnum = None
        self.starting_port = starting_port
        self.url = ""
        self._find_portnum()
        self.tasks = {}
        self.metadata = {}
        logger.info("New Machinedata: %s:%s" % (self.hostname,
                                                self.portnum))

    def _find_portnum(self):
        found = False
        port = self.starting_port
        while not found:
            try:
                logger.info("Attempting to connect to %s" % (
                        "%s:%s" % (self.hostname, port)))

                # Verify we can actually make an http request
                req = requests.get("http://%s:%s" % (self.hostname,
                                                     port),
                                   timeout=2)
                if not req.content:
                    raise Exception("No content recieved!")

                found = True
                self.portnum = port
                logger.info("Successfully connected to %s:%s" % (
                        self.hostname, self.portnum))
            except:
                import traceback
                logger.warn("Failed to connect to %s:%s" % (
                        self.hostname, port))
                logger.warn(traceback.format_exc().splitlines()[-1])
                port += 1
                if port > self.starting_port + 15:
                    self.url = ""
                    self.portnum = None
                    return

        self.url = "http://%s:%s" % (self.hostname,
                                     self.portnum)
        return self.url

    def _make_request(self, function, path, host=None):
        val = None
        hostname = host
        if not hostname:
            hostname = self.url

        try:
            val = function("%s/%s" % (hostname, path),
                           timeout=5)
        except:
            self._find_portnum()
            hostname = host
            if not hostname:
                hostname = self.url
            try:
                val = function("%s/%s" % (hostname, path),
                               timeout=5)
            except:
                logger.warn("Couldn't execute %s/%s!" % (
                        hostname, path))
                import traceback
                logger.error(traceback.format_exc())
                return None

        return val

    def load_generic_page(self, host, page):
        response = self._make_request(requests.get,
                                      path="%s?nohtml=1&format=json" % page,
                                      host=host)
        if not response:
            return {}

        data = simplejson.loads(response.content)
        return data

    def reload(self):
        response = self._make_request(requests.get,
                                      path="stats?nohtml=1&format=json")
        if not response:
            return None

        data = simplejson.loads(response.content)

        task_data = {}
        new_tasks = {}
        for key, value in data.iteritems():
            try:
                task_name, metric = key.split('-')
            except:
                self.metadata[key] = value
                continue

            if not task_name in task_data:
                task_data[task_name] = {}
            task_data[task_name][metric] = value

        for task_name in task_data.keys():
            task_dict = task_data[task_name]
            new_tasks[task_dict['name']] = task_dict
            if task_dict['running']:
                updated = False
                tries = 0
                while not updated and tries < 10:
                    tries += 1
                    try:
                        new_tasks = self.update_task_data(new_tasks,
                                                         task_dict['name'])
                        updated = True
                    except:
                        # Http server might not be up yet
                        time.sleep(0.01)
                        logger.warn("Couldn't update task")

        self.tasks = new_tasks
        return self.tasks

    def update_task_data(self, tasks, task_name):
        stats_page = self.strip_html(tasks[task_name]['monitoring'])
        tasks[task_name].update(
            self.load_generic_page(
                stats_page,
                'stats'))
        tasks[task_name]['stats_page'] = "%s/stats" % stats_page
        tasks[task_name]['logs_page'] = "%s/logs" % stats_page
        return tasks

    def add_task(self, config):
        params = '&'.join(
            "%s=%s" % (
                k, urllib.quote_plus(str(v))) for k, v in config.items())
        val = self._make_request(requests.get,
                                 path="add_task?%s" % params)
        if val:
            return val.content
        else:
            return None

    def start_task(self, task):
        if isinstance(task, str):
            task = self.tasks[task]

        tid = urllib.quote(task['name'])
        val = self._make_request(
            requests.get,
            path="start_task?task_name=%s" % tid)

        if val:
            return val.content

        return None

    def stop_task(self, task):
        if isinstance(task, str):
            task = self.tasks[task]

        tid = urllib.quote(task['name'])
        val = self._make_request(
            requests.get,
            path="stop_task?task_name=%s" % tid)

        if val:
            return val.content

        return None

    def strip_html(self, val):
        return re.sub('<[^<]+?>', '', val)

    def get_sitter_logs(self):
        return self.load_generic_page(self.url, "logs")

    def get_task_logs(self, task):
        sitter_logs = self.get_sitter_logs()
        logs = {}
        for logname, logfile in sitter_logs.items():
            if task['name'] in logname:
                logs[logname] = logfile
        try:
            url = self.strip_html(task['monitoring'])
            logs.update(self.load_generic_page(url, 'logs'))
        except:
            pass

        return logs

    def get_logfile(self, task, stderr=False):
        url = self.strip_html(task['monitoring'])
        data = self.load_generic_page(url,
                                      'stats')
        logs = self.load_generic_page(url,
                                      'logs')

        tasknum = str(int(data['num_task_starts']) - 1)
        handle = "stdout.%s" % tasknum
        if stderr:
            handle = "stderr.%s" % tasknum

        return logs[handle]

if __name__ == '__main__':
    import subprocess
    d = MachineData("localhost", 40000)
    d.reload()
    print d.tasks
    task = d.tasks['REX (Remote Extractor)']
    print d.get_task_logs(task)



import logging
import os
import socket
import sys
import threading
from datetime import datetime
from logging import FileHandler

import clusterstate
import deploymentrecipe
import dynect
import eventmanager
import jobfiller
import machinedoctor
import machinemonitor
import monitoredmachine
import productionjob
import providers.aws
import sittercommon.machinedata


from clusterstate import ClusterState
from clusterstats import ClusterStats
from eventmanager import ClusterEventManager
from machinedoctor import MachineDoctor
from machinemonitor import MachineMonitor
from monitoredmachine import MonitoredMachine
from productionjob import ProductionJob
from providers.aws import AmazonEC2
from sittercommon import http_monitor
from sittercommon import logmanager

logger = logging.getLogger(__name__)

"""
In an ideal world this is written such that we always speak
in terms of CPUs and RAM requirements for each job.  This allows
us to put multiple jobs on a machine whilst keeping track
of total theoretical machine utilization.  However, for simplicities
sake we're cutting some corners (FOR NOW) and assuming one job
per machine.  These sections are noted with the comment
#!MACHINEASSUMPTION!
"""


class ClusterSitter(object):
    def __init__(self, log_location, daemon,
                 provider_config,
                 dns_provider_config,
                 keys=None, user=None,
                 starting_port=30000,
                 launch_location=None):
        self.worker_thread_count = 4
        self.daemon = daemon
        self.keys = keys
        self.user = user
        self.provider_config = provider_config
        self.dns_provider_config = dns_provider_config
        self.log_location = log_location
        self.launch_location = launch_location
        self.launch_time = datetime.now()
        self.start_state = "Not Started"

        self.state = ClusterState(self)

        self.orig_starting_port = starting_port
        self.next_port = starting_port
        self.start_count = 1
        self.command = "clustersitter"
        self.parent_pid = os.getpid()
        self.logmanager = logmanager.LogManager(stdout_location=log_location,
                                                stderr_location=log_location)
        self.logmanager.set_harness(self)

        self.dns_provider = None
        if 'class' in dns_provider_config:
            (module, clsname) = dns_provider_config['class'].split(':')
            clz = self._get_recipe_class(module, find_func=clsname)
            self.dns_provider = getattr(clz, clsname)(
                dns_provider_config)

        # In seconds
        self.stats_poll_interval = 5

        self.stats = ClusterStats(self)
        self.http_monitor = http_monitor.HTTPMonitor(self.stats,
                                                     self,
                                                     30000)

        self.http_monitor.add_handler('/overview', self.stats.overview)
        self.http_monitor.add_handler('/add_job', self.api_add_job)
        self.http_monitor.add_handler('/remove_job', self.api_remove_job)
        self.http_monitor.add_handler('/update_idle_limit',
                                      self.api_enforce_idle)
        self.http_monitor.add_handler('/update_logging_level',
                                      self.api_update_logging_level)
        self.http_monitor.add_handler('/update_job',
                                      self.api_update_job)

        # Do lots of logging configuration
        modules = [sys.modules[__name__],
                   machinemonitor,
                   monitoredmachine,
                   productionjob,
                   providers.aws,
                   deploymentrecipe,
                   sittercommon.machinedata,
                   jobfiller,
                   eventmanager,
                   clusterstate,
                   machinedoctor,
                   dynect]

        formatter = logging.Formatter(
            '%(asctime)s - %(name)s:%(threadName)s'
            '.%(levelname)s - %(message)s')

        if not os.path.exists(self.log_location):
            os.makedirs(self.log_location)

        all_file = FileHandler("%s/all.log" % self.log_location)
        all_file.setFormatter(formatter)

        self.logfiles = ["%s/all.log" % self.log_location]

        for module in modules:
            name = module.__name__.split('.')[-1]
            logfile = "%s/%s.log" % (self.log_location,
                                                 name)
            self.logfiles.append(logfile)
            self.logmanager.add_logfile(name, logfile)
            handler = FileHandler(logfile)
            handler.setFormatter(formatter)
            module.logger.addHandler(all_file)
            module.logger.addHandler(handler)
            self.state.loggers.append(module.logger)

        socket.setdefaulttimeout(2)

    # ----------- API ----------------
    def _api_check(self, args, required_fields):
        if self.start_state != "Started":
            return "Not ready to recieve API calls"

        for field in required_fields:
            if not field in args:
                return "Missing Field %s" % field

        return

    def api_update_logging_level(self, args):
        # Default level is INFO
        level = args.get('level', 20)
        for logger in self.state.loggers:
            logger.setLevel(level)

        ClusterEventManager.handle(
                'Updated logging level to %s' % level)

        return "Level set to %s" % level

    def api_enforce_idle(self, args):
        # Really naive right now, a global # of
        # max idle per zone.  Could do a lot more here.
        check = self._api_check(args, ['idle_count_per_zone'])

        if check:
            return check

        try:
            self.state.max_idle_per_zone = int(args['idle_count_per_zone'])
        except:
            return "Invalid limit"

        ClusterEventManager.handle(
                'Enforce Idle Limit at %s' % int(args['idle_count_per_zone']))
        return "Limit set"

    def api_update_job(self, args):
        check = self._api_check(args,
                                ['job_name'])

        if check:
            return check

        job_name = args['job_name']
        job = None
        for state_job in self.state.jobs:
            if state_job.name == job_name:
                job = state_job
                break

        if not job:
            return "Error updating job: %s doesn't exist" % job_name

        job.do_update_deployment(self.state, args.get('version'))
        # Now build a deployment recipe for this job
        ClusterEventManager.handle(
            'Update %s started' % job_name)

        return "Job update initiated"

    def api_add_job(self, args):
        check = self._api_check(args,
                                ['dns_basename',
                                 'task_configuration',
                                 'deployment_layout',
                                 'deployment_recipe',
                                 'recipe_options',
                                 'persistent'])
        if check:
            return check
        job = ProductionJob(
                args['dns_basename'],
                args['task_configuration'],
                args['deployment_layout'],
                args['deployment_recipe'],
                args['recipe_options'],
                args['persistent'],
                args.get('linked_job'))

        if args.get('linked_job'):
            job.find_linked_job(self.state)
            if not job.linked_job_object:
                return "Couldn't find linked job!"

        if self.state.add_job(job):
            ClusterEventManager.handle(
                "Added a job: %s" % job.get_name())

            return "Job Added"
        else:
            return "Error adding job, see logs"

    def api_remove_job(self, args):
        check = self._api_check(args, ['name'])

        if check:
            return check

        if self.state.remove_job(args['name']):
            ClusterEventManager.handle(
                "Removed a job: %s" % args['name'])

            return "Removal OK"
        else:
            return "Couldn't find job to remove"

    # ----------- END API ----------

    def machines_in_queue(self):
        for monitor, thread in self.state.monitors:
            if monitor.processing_new_machines():
                return True

        return False

    def _get_recipe_class(self, recipe_class, find_func="run_deploy"):
        if isinstance(recipe_class, type):
            return recipe_class

        recipe_cls = None
        if isinstance(recipe_class, basestring):
            try:
                recipe_cls = __import__(recipe_class,
                                          globals(),
                                          locals())
            except:
                # Odd?
                logger.warn("Not sure what %s is..." % recipe_class)
        else:
            logger.warn("Not sure what %s is..." % recipe_class)
            return None

        if hasattr(recipe_cls, find_func):
            return recipe_cls

        exports = dir(recipe_cls)
        for export in exports:
            if hasattr(getattr(recipe_cls, export), find_func):
                return getattr(recipe_cls, export)

        try:
            module = sys.modules[recipe_class]
        except:
            return None

        exports = dir(module)
        for export in exports:
            if hasattr(getattr(module, export), find_func):
                value = getattr(module, export)
                if value.__module__ == module.__name__:
                    return value

    def build_recipe(self, recipe_class, machine,
                     post_callback=None, options=None,
                     given_logger=None):
        username = self.user
        keys = self.keys
        if machine.config.login_name:
            username = machine.config.login_name

        if machine.config.login_key:
            keys.append(machine.config.login_key)

        recipe_cls = self._get_recipe_class(recipe_class)
        if not recipe_cls:
            return None

        return recipe_cls(machine.config.hostname,
                          username,
                          keys,
                          post_callback=post_callback,
                          options=options,
                          given_logger=given_logger,
                          dns_hostname=machine.config.dns_name,
                          launch_location=self.launch_location)

    def get_next_port(self):
        works = None
        while not works:
            try:
                sok = socket.socket(socket.AF_INET)
                sok.bind(("0.0.0.0", self.next_port))
                sok.close()
                works = self.next_port
                self.next_port += 1
            except:
                self.next_port += 1
                if self.next_port > self.orig_starting_port + 10000:
                    self.next_port = self.orig_starting_port
        return works

    def remove_machine(self, machine):
        self.state.machines_by_zone[
            machine.config.shared_fate_zone].remove(machine)

        for monitor in self.state.monitors:
            monitor[0].remove_machine(machine)

    def decomission_machine(self, machine):
        self.remove_machine(machine)
        if not machine.config.shared_fate_zone in self.state.provider_by_zone:
            logger.warn(
                "No provider found for %s?" % machine.config.shared_fate_zone)
            return

        provider = self.state.provider_by_zone[machine.config.shared_fate_zone]

        ClusterEventManager.handle(
            "Decomissioning %s" % str(machine))

        if not provider.decomission(machine):
            # If we can't decomission it then perhaps its locked
            # and we should leave well enough alone at this point,
            # just remove it from monitoring etc.
            return

        if machine.config.dns_name:
            self.dns_provider.remove_record(data=machine.config.ip,
                                            hostName=machine.config.dns_name)

            # Strip off the leading number, e.g.
            # 12.bar.mydomain.com -> bar.mydomain.com
            root_name = '.'.join(machine.config.dns_name.split('.')[1:])

            self.dns_provider.remove_record(data=machine.config.ip,
                                            hostName=root_name)

        # Now look for other dangling records pointing to this machine
        # and delete those too.
        records = self.dns_provider.get_records()
        for record in records:
            if record['value'] == machine.config.ip:
                logger.info("Removing %s from %s" % (
                        machine.config.ip,
                        record['record']))
                self.dns_provider.remove_record(data=machine.config.ip,
                                                hostName=record['record'])

        ClusterEventManager.handle(
            "Decomissioning of %s complete!" % str(machine))

    def add_machines(self, machines, update_dns=True):
        if not machines:
            return

        monitored_machines = []
        logger.info("Adding machines to monitoring: %s" % machines)
        for m in machines:
            mm = m
            if not isinstance(mm, MonitoredMachine):
                mm = MonitoredMachine(m)

            if not mm.config.shared_fate_zone in self.state.machines_by_zone:
                self.state.machines_by_zone[mm.config.shared_fate_zone] = []

            # We can re-add machines after they temporarily disappear
            if mm not in self.state.machines_by_zone[
                mm.config.shared_fate_zone]:
                self.state.machines_by_zone[
                    mm.config.shared_fate_zone].append(mm)

            monitored_machines.append(mm)

        # Spread the machines out evenly across threads
        num_threads_to_use = min(self.worker_thread_count, len(machines)) or 1
        machines_per_thread = int(len(machines) / num_threads_to_use) or 1

        # Take care of individisable # of machines by thread
        if machines_per_thread * num_threads_to_use != len(machines):
            machines_per_thread += 1

        self.state.monitors.sort(key=lambda x: x[0].num_monitored_machines())
        monitors_to_use = self.state.monitors[:num_threads_to_use]

        logger.info(
            "Add to Monitoring, num_threads_to_use: %s" % num_threads_to_use +
            "machines_per_thread: %s, " % machines_per_thread)

        for index, monitor in enumerate(monitors_to_use):
            start_index = index * machines_per_thread
            end_index = (index + 1) * machines_per_thread

            if end_index > len(machines):
                end_index = len(machines)

            if start_index > len(machines):
                start_index = end_index

            if start_index != end_index:
                monitor[0].add_machines(
                    monitored_machines[start_index:end_index])

        if not update_dns:
            return

        # Ensure we have up to date DNS names for each machine
        logger.info("Loading DNS Records...")

        records = self.dns_provider.get_records()
        record_by_ip = {}
        for record in records:
            record_by_ip[record['value']] = record['record']

        for machine in monitored_machines:
            if not machine.config.dns_name:
                ip = socket.gethostbyname(machine.config.hostname)
                if ip in record_by_ip:
                    machine.config.dns_name = record_by_ip[ip]
                    logger.info("Found name %s for %s" % (
                            machine.config.dns_name,
                            ip))

    def start(self):
        self.http_monitor.start()
        logger.info("Cluster Sitter Monitor started at " + \
            "http://localhost:%s" % self.http_monitor.port)

        self.start_state = "Starting Up"

        logger.info("Initializing MachineProviders")
        # Initialize (non-started) machine monitors first
        # Then download machine data, populate data structure and monitors
        # then kickoff everything.

        logger.info(
            "Initializing %s MachineMonitors" % self.worker_thread_count)
        # Spin up all the monitoring threads
        for threadnum in range(self.worker_thread_count):
            machinemonitor = MachineMonitor(parent=self,
                                            number=threadnum)
            thread = threading.Thread(target=machinemonitor.start,
                                      name='Monitoring-%s' % threadnum)
            self.state.monitors.append((machinemonitor, thread))

        aws = AmazonEC2(self.provider_config['aws'])
        if aws.usable():
            self.state.providers['aws'] = aws

        for provider in self.state.providers.values():
            newzones = provider.get_all_shared_fate_zones()
            self.state.zones.extend(newzones)
            for zone in newzones:
                self.state.provider_by_zone[zone] = provider

            # Note: add_machines() has to be called AFTER the monitors
            # are initialized.
            self.add_machines(provider.get_machine_list())

        logger.info("Zone List: %s" % self.state.zones)

        # Kickoff all the threads at once
        if self.daemon:
            self.logmanager.setup_all()

        logger.info("Spinning up %s monitoring threads" % (
                self.worker_thread_count))
        for monitor in self.state.monitors:
            monitor[1].start()

        logger.info("Starting metadata calculator")
        self.calculator = threading.Thread(target=self.state._calculator,
                                           name="Calculator")
        self.calculator.start()

        logger.info("Starting machine recovery thread")
        self.machine_doctor = MachineDoctor(self)
        self.machine_doctor.start()

        self.start_state = "Started"

    def _register_sitter_failure(self, monitored_machine, monitor):
        """
        !! Fabric is not threadsafe.  Do all work in one thread by appending to
        a queue !!
        # Update -- not using fabric anymore, but we still don't want
        to be mucking with machines in the machinemonitor thread,
        still a good idea to shift to machinedoctor

        """
        logger.info(
            "Registering an unreachable machine %s" % monitored_machine)
        self.state.unreachable_machines.append((monitored_machine, monitor))



"""
A class which encapsulates a process and a set of constraints
and ensures that they are constantly fulfilled.
"""
import datetime
import os
import pwd
import signal
import simplejson
import sys
import threading
import time

import process


class ProcessHarness(object):
    """
    An object which manages the lifecycle of a single child process, killing it
    when it violates constraitns and rebooting it as necessary
    """
    def __init__(self, command, constraints, restart=False,
                 max_restarts=-1, poll_interval=.1,
                 logmanager=None, uid=None, allow_spam=False,
                 collect_stats=True):
        self.launch_location = os.getcwd()
        self.child_proc = None
        self.child_running = True
        self.collect_stats = collect_stats
        self.command = command
        self.constraints = constraints
        self.max_restarts = max_restarts
        self.poll_interval = poll_interval
        self.restart = restart
        self.start_count = 0
        self.uid = uid
        self.parent_pid = os.getpid()
        self.logmanager = logmanager
        self.logmanager.set_harness(self)
        self.stop_running = False
        self.last_start = datetime.datetime.min
        self.allow_spam = allow_spam
        # Statistics
        self.task_start = datetime.datetime.now()
        self.violations = {}
        for constraint in self.constraints:
            self.violations[str(constraint)] = 0

        signal.signal(signal.SIGTERM, self.exit_now)
        signal.signal(signal.SIGINT, self.exit_now)

        if self.uid != None:
            actual_uid = None
            try:
                actual_uid = int(self.uid)
            except:
                try:
                    import pwd
                    actual_uid = pwd.getpwnam(self.uid)[2]
                except:
                    sys.stderr.write("Invalid UID: %s" % uid)
                    os._exit(1)

            self.uid = actual_uid

        # Start the child process
        self.start_process()

    def exit_now(self, *_):
        self.stop_running = True
        self.child_running = False
        self.terminate_child()
        print simplejson.dumps(self.logmanager.get_logfile_names())
        os._exit(0)

    def start_process(self):
        """
        Start a new instance of the child task
        """
        # Avoid spam-restarts, only allow restarting
        # once per second
        now = datetime.datetime.now()
        if not self.allow_spam and \
                now - self.last_start < datetime.timedelta(seconds=1):
            time.sleep(1)

        self.last_start = datetime.datetime.now()

        pid = os.fork()
        if pid == 0:
            # We're the child, we'll exec
            # Put ourselves into our own pgrp, for sanity
            os.setpgrp()

            # Configure STDOUT and STDERR
            self.logmanager.setup_stdout()
            self.logmanager.setup_stderr()

            if self.uid != None:
                try:
                    os.setuid(self.uid)
                except OSError:  # no permission
                    sys.stderr.write("Must be root to set UID!")
                    os._exit(1)

                # Now that we're the child and the child's PID setup
                # the environment for child scripts.
                userpwd = pwd.getpwuid(self.uid)
                os.environ['HOME'] = userpwd.pw_dir
                os.environ['USER'] = userpwd.pw_name

            # parse the command
            cmd = '/bin/bash'
            args = [cmd, "-c", self.command]
            os.execvp(cmd, args)

        self.child_proc = process.Process(pid)
        self.start_count += 1

    def do_monitoring(self):
        """
        Begin monitoring the child process
        """
        while True:
            if self.stop_running:
                return

            if self.collect_stats:
                self.child_proc.update_usage(deep=True)

            restarted = False

            for constraint in self.constraints:
                if constraint.check_violation(self.child_proc):
                    if self.child_violation_occured(constraint):
                        print "Restarting child command %s" % self.command
                        self.start_process()
                        restarted = True

            if not self.child_proc.is_alive():
                # The child proc could have died inbetween checking
                # constraints and now.  If there is a LivingConstraint
                # then fire it
                for constraint in self.constraints:
                    if str(constraint) == "LivingConstraint":
                        if self.child_violation_occured(constraint):
                            print "Restarting child command %s" % self.command
                            self.start_process()
                            restarted = True

                # If we restarted the child proc we don't want to set
                # child_running to False because... its True =P
                if not restarted:
                    # There is no living constraint and child is dead,
                    # so set running to false
                    print "Child exited on its own, not asked to " + \
                        "restart it, exiting"
                    self.child_running = False
                    return

            time.sleep(self.poll_interval)

    def child_violation_occured(self, violated_constraint):
        """
        Take appropriate action when we're in violation
        Returns:
          True means the process should be restarted
          False means the caller should take no action
        """
        print "Violated Constraint %s" % str(violated_constraint)
        if violated_constraint.kill_on_violation:
            self.child_proc.force_exit()

        self.violations[str(violated_constraint)] += 1

        if self.restart:
            if (self.max_restarts == -1 or
                self.start_count <= self.max_restarts):
                return True

        self.child_running = False
        return False

    def begin_monitoring(self):
        """Split off a thread to monitor the child process"""
        monitoring_thread = threading.Thread(target=self.do_monitoring,
                                             name='Child Monitoring')

        monitoring_thread.start()

    def terminate_child(self):
        """
        Kill the child process
        """
        self.child_proc.force_exit()
        self.wait_for_child_to_finish()

    def wait_for_child_to_finish(self):
        """
        Wait for the child process to complete naturally.
        """
        code = 0
        while self.child_running:
            _, newcode = self.child_proc.wait_for_completion()
            if newcode:
                code = newcode

            print "Child %s exited %s" % (self.child_proc.pid, code)
            time.sleep(.1)

        return code



import json
import requests

from sittercommon import arg_parser


def get_help_string():
    return "Update clustersitter idle machine limit"


def get_command():
    return "update_idle_limit"


def get_parser(parser):
    parser.add_argument("--idle-limit", dest="idle_limit",
                        required=True,
                        help='Maximum number of idle machines per zone')

    return parser


def run_command(clustersitter_url, idle_limit):
    data = {'data': json.dumps({'idle_count_per_zone': idle_limit})}
    print "%s/update_idle_limit" % clustersitter_url
    resp = requests.post("%s/update_idle_limit" % clustersitter_url,
                         data=data)
    print resp.content



"""
Manage the various logging facilities for a child process.
"""
import md5
import os
import random


class LogManager(object):
    """
    An object to manage logging facilities for the child
    """
    def __init__(self, stdout_location='-', stderr_location='-'):
        self.stdout_location = stdout_location
        self.stderr_location = stderr_location

        # Enable others to write to the log locations incase the child
        # Process is run as another user.
        os.system("chmod o+rwx %s" % stdout_location)
        os.system("chmod o+rwx %s" % stderr_location)
        self.harness = None
        self.extra_logfiles = {}

    def set_harness(self, harness):
        self.harness = harness

    def add_logfile(self, name, filename):
        self.extra_logfiles[name] = filename

    def setup_all(self):
        stdout_loc = self._calculate_filename(
            self.stdout_location)
        stderr_loc = self._calculate_filename(
            self.stderr_location, True)
        print "Redirecting sitter output to %s, stderr: %s" % (
            stdout_loc, stderr_loc)

        self.add_logfile("stdout", stdout_loc)
        self.add_logfile("stderr", stderr_loc)

        self.setup_stdout()
        self.setup_stderr()

    def setup_stdout(self):
        if self.stdout_location != '-':
            stdout = open(self._calculate_filename(self.stdout_location),
                          'w')
            stdout_fileno = stdout.fileno()
            os.dup2(stdout_fileno, 1)

    def setup_stderr(self):
        if self.stderr_location != '-':
            stderr = open(self._calculate_filename(self.stderr_location,
                                                   True),
                          'w')
            stderr_fileno = stderr.fileno()
            os.dup2(stderr_fileno, 2)

    def get_logfile_names(self):
        filenames = {}
        for i in range(self.harness.start_count):
            if self.stdout_location != "-":
                filenames["stdout.%d" % i] = self._calculate_filename(
                    self.stdout_location,
                    False,
                    str(i))

            if self.stderr_location != "-":
                filenames["stderr.%d" % i] = self._calculate_filename(
                    self.stderr_location,
                    True,
                    str(i))

        filenames.update(self.extra_logfiles)
        return filenames

    def _calculate_filename(self, directory, stderr=False,
                            number=None):
        if not os.path.exists(directory):
            os.makedirs(directory)

        filenum = number
        if not number and self.harness:
            filenum = self.harness.start_count

        parent_pid = os.getpid()

        if self.harness:
            payload = self.harness.command
            parent_pid = self.harness.parent_pid
        else:
            payload = str(random.random())

        payload += str(parent_pid)

        if stderr:
            payload += "err"

        name = "%s/%s.%s" % (directory,
                             md5.md5(payload).hexdigest(),
                             filenum)
        return name



import curses
import os
import socket
import subprocess
import sys

from datetime import datetime
from menu import MenuFactory, MenuOption, MenuChanger, Table
from sittercommon.machinedata import MachineData

MACHINE_DATA = None
MENUFACTORY = None
SCR = None
YPOS = 0
CURRENT_LOC = "mainmenu"
AUX = None


def add_line(msg):
    global YPOS
    SCR.addstr(YPOS, 0, msg)
    YPOS += 1


def remove_line():
    global YPOS
    SCR.clrtoeol()
    y, x = SCR.getyx()
    SCR.move(y - 1, x)
    YPOS -= 1


def refresh():
    global YPOS
    SCR.clear()
    SCR.refresh()
    YPOS = 0


def header():
    add_line("#" * SCR.getmaxyx()[1])
    add_line(
        "MachineSitter at %s - Curses UI %s" % (
            MACHINE_DATA.url, datetime.now()))
    add_line("#" * SCR.getmaxyx()[1])


def change_menu(newmenu, aux=None):
    global CURRENT_LOC, AUX
    CURRENT_LOC = newmenu
    AUX = aux


def stop_sitter():
    os.kill(int(MACHINE_DATA.metadata['machinesitter_pid']), 15)
    sys.exit(0)


def mainmenu():
    menu = MENUFACTORY.new_menu("Main Menu")
    menu.add_option_vals("Refresh Window", action=dir, hotkey="*")
    menu.add_option_vals("Add a new task",
                    action=lambda: change_menu('addtask'))

    menu.add_option_vals("Show task definitions",
                    action=lambda: change_menu('show_task_definitions'))

    menu.add_option_vals("Show machine sitter logs",
                    action=lambda: change_menu('show_machinesitter_logs'))

    menu.add_option_vals("Stop machine sitter",
                         action=stop_sitter)

    menu.render()


def show_machinesitter_logs():
    logs = MACHINE_DATA.get_sitter_logs()
    show_logs(logs, "Machine Sitter Logs")


def show_task_logs():
    task = AUX
    logs = MACHINE_DATA.get_task_logs(task)
    show_logs(logs, "%s Logs" % task['name'])


def show_logs(logs, title):
    menu = MENUFACTORY.new_menu(title)
    menu.add_option_vals("Main Menu",
                    action=lambda: change_menu('mainmenu'), hotkey="*")

    lognames = logs.keys()
    lognames.sort()

    for logname in lognames:
        logfile = logs[logname]['location']
        menu.add_option_vals("%s (%s)" % (logname, logfile),
                         action=MenuChanger(tail_file, [logfile]))

    menu.render()


def tail_file(filenames):
    curses.endwin()
    os.system("clear")
    subprocess.call(["echo", ' '.join(filenames)])
    try:
        args = ["tail", "-n", "100", "-f"]
        args.extend(filenames)
        subprocess.call(args)
    except:
        pass


def show_log(task):
    tail_file([
            MACHINE_DATA.get_logfile(task)['location'],
            MACHINE_DATA.get_logfile(task, True)['location']])


def start_task(task):
    MACHINE_DATA.start_task(task)


def stop_task(task):
    MACHINE_DATA.stop_task(task)


def show_task():
    name, task = AUX
    reload_data()
    task = MACHINE_DATA.tasks[name]

    menu = MENUFACTORY.new_menu("%s (%s) (%s)" % (
            name,
            task['command'],
            MACHINE_DATA.strip_html(task.get('monitoring', 'Not Running'))
            ))

    menu.add_option_vals("Main Menu",
                    action=lambda: change_menu('mainmenu'), hotkey="*")

    if task['running'] == "False":
        menu.add_option_vals("Start Task",
                             action=lambda: start_task(task))
    else:
        menu.add_option_vals("Stop Task",
                             action=lambda: stop_task(task))

        menu.add_option_vals("Show stdout/stderr",
                         action=lambda: show_log(task))

    menu.add_option_vals("Show historic task log files",
                         action=lambda: change_menu('show_task_logs', task))

    menu.render()


def basic_tasks():
    global MENUFACTORY
    add_line("-" * SCR.getmaxyx()[1])

    reload_data()

    running = []
    not_running = []
    not_running_lines = []
    MENUFACTORY = MenuFactory(SCR, add_line, remove_line)

    for name, task in MACHINE_DATA.tasks.items():
        if task['running']:
            running.append((name, task))
        else:
            not_running.append((name, task))

    hotkey = 1

    table = Table(SCR.getmaxyx()[1],
                  ["Running Task Name", "Restarts",
                   "Runtime", "Stdout", "Stderr"])

    for name, task in running:
        runtime = '?'
        stdout_kb = -1.0
        stderr_kb = -1.0
        if task.get('process_start_time'):
            runtime = str(
                datetime.now() - datetime.strptime(
                    task['process_start_time'],
                    '%Y-%m-%d %H:%M:%S.%f')
                )

            # strip useconds
            runtime = runtime[:runtime.find('.')]
            try:
                stdout = MACHINE_DATA.get_logfile(task)['location']
                stderr = MACHINE_DATA.get_logfile(task, True)['location']

                stdout_kb = os.stat(stdout).st_size / 1024
                stderr_kb = os.stat(stderr).st_size / 1024
            except:
                import traceback
                traceback.print_exc()
                pass

        table.add_row([task['name'],
                       task.get('num_task_starts', '?'),
                       runtime,
                       "%.0f kB" % stdout_kb,
                       "%.0f kB" % stderr_kb])

        option = MenuOption(
                task['name'],
                action=MenuChanger(change_menu, "show_task",
                                   (name, task)),
                hotkey=str(hotkey),
                hidden=True)

        MENUFACTORY.add_default_option(option)
        hotkey += 1

    for name, task in not_running:
        line = task['name']
        not_running_lines.append(line)
        option = MenuOption(
                task['name'],
                action=MenuChanger(change_menu, "show_task",
                                   (name, task)),
                hotkey=str(hotkey),
                hidden=True)

        MENUFACTORY.add_default_option(option)
        hotkey += 1

    show_table = str(table).split('\n')
    add_line("   %s" % show_table[0])

    for num, l in enumerate(show_table[1:]):
        add_line("%s. %s" % ((num + 1), l))

    add_line("Stopped Tasks:")
    for num, l in enumerate(not_running_lines):
        add_line("%s. %s" % ((len(running) + num + 1), l))

    add_line("-" * SCR.getmaxyx()[1])


def reload_data():
    global MACHINE_DATA
    MACHINE_DATA.reload()


def run():
    global SCR, MENUFACTORY
    SCR = curses.initscr()

    refresh()
    reload_data()

    while True:
        refresh()
        header()
        basic_tasks()
        globals()[CURRENT_LOC]()


def main():
    global MACHINE_DATA
    if not MACHINE_DATA:
        MACHINE_DATA = MachineData("localhost", 40000)

    try:

        if not MACHINE_DATA.url:
            print "Couldn't find a running machine sitter!"
            return

        run()
    except:
        curses.endwin()
        import traceback
        traceback.print_exc()

if __name__ == '__main__':
    main()



"""
Reads data from a stats collector and exposes it via HTTP
"""
import cgi
import os
import simplejson
import threading
import tenjin
import urlparse
import sys
import SocketServer
import BaseHTTPServer
from pkg_resources import resource_filename

# A weird requirement from tenjin to have this
# The world explodes if we don't have it
from tenjin.helpers import *


def head(filename, num_lines):
    return file_process("head", filename, num_lines)


def tail(filename, num_lines):
    return file_process("tail", filename, num_lines)


def file_process(cmd, filename, num_lines):
    """
    Implementing tail in python is like 50 lines of code,
    and is pretty unreadable.  Just cop out and use something
    we know works.

    """
    stdin, stdout = os.popen2("%s -n %s %s" % (cmd, num_lines, filename))
    stdin.close()
    data = stdout.read()
    stdout.close()
    return data


class HTTPMonitorHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    def __init__(self, monitor, new_handlers, *args, **kwargs):
        self.monitor = monitor

        self.handlers = {
            "/stats": self._get_stats,
            "/logs": self._get_logs,
            "/logfile": self._get_logfile,
        }

        self.handlers.update(new_handlers)

        paths = [
            resource_filename(__name__, 'templates'),
            'templates',
            '/usr/local/cerebro/templates',
            '/opt/tasksitter/templates/',
            os.path.join(
                os.getenv('HOME'),
                'workspace',
                'tasksitter',
                'templates')]

        if hasattr(monitor.harness, "launch_location"):
            paths.append(
                os.path.join(monitor.harness.launch_location, "templates"))

        tenjin.Engine.cache = tenjin.MemoryCacheStorage()
        self.engine = tenjin.Engine(path=paths)

        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def _usage(self, _):
        output = """Invalid Path Requests.  Options:<br><ul>"""

        return self.engine.render('index.html', {'handlers':
                                                 self.handlers,
                                                 'name': sys.argv[0]})

        for handler in self.handlers.keys():
            output += "<li><a href='%s'>%s</a></li>" % (handler, handler)
        output += "</ul></body></html>"
        return output

    def add_handler(self, path, callback):
        self.handlers[path] = callback

    def _get_logfile(self, args):
        # @TODO This is a bit of a security problem...
        filename = args.get('name')
        if not filename:
            if not 'logname' in args:
                return "No filename specified"

            logfiles = self.monitor.get_logs()
            filename = logfiles[args['logname']]

        filehandle = None
        try:
            filehandle = open(filename)
        except IOError:
            return "File not found"

        if not args.get('head') and not args.get('tail'):
            return filehandle.read()

        if args.get('tail'):
            return tail(filename, int(args['tail']))
        else:
            return head(filename, int(args['head']))

    def _get_logs(self, args):
        logfiles = self.monitor.get_logs()
        for k, v in logfiles.items():
            size = 0
            try:
                size = float(os.stat(v).st_size) / 1024
            except:
                pass

            logfiles[k] = {'url': "/logfile?name=%s" % v,
                           'location': v,
                           'size': size}

        if 'nohtml' in args:
            return self._format_dict(logfiles, args)
        else:
            return self.engine.render('logs.html', {'data': logfiles})

    def _get_stats(self, args):
        stats = self.monitor.get_stats()
        if "nohtml" in args:
            return self._format_dict(stats, args)
        else:
            return self.engine.render('stats.html', {'data': stats})

    def _format_dict(self, data, args):
        """
        Convert a dictionary of data into an appropriate
        output format based on the query string args
        """
        def todict(obj):
            if isinstance(obj, dict):
                return obj

            try:
                return obj.__dict__
            except:
                return str(obj)

        output = ""
        if "format" in args and args["format"] != "flat":
            if args['format'] == "json":
                output = simplejson.dumps(data,
                                          skipkeys=True,
                                          default=todict)
            else:
                output = "Invalid Format"
        else:
            seperator = "\n"
            for key, value in data.items():
                output += "%s=%s%s" % (key, value, seperator)

        return output

    def do_GET(self):
        urldata = urlparse.urlparse(self.path)
        array_args = urlparse.parse_qs(urldata.query)
        args = dict([(k, v[0]) for k, v in array_args.items()])

        if not urldata.path in self.handlers:
            self.wfile.write(self._usage(args))
            return

        args['engine'] = self.engine

        try:
            output = self.handlers[urldata.path](args)
            if not output:
                output = "No Data"

            if isinstance(output, dict):
                self.wfile.write(self._format_dict(output, args))
            else:
                self.wfile.write(output)
        except:
            import traceback
            self.wfile.write(traceback.format_exc())

        return

    def do_POST(self):
        urldata = urlparse.urlparse(self.path)
        ctype, pdict = cgi.parse_header(self.headers.getheader('content-type'))
        if ctype == 'application/x-www-form-urlencoded':
            length = int(self.headers.getheader('content-length'))
            try:
                args = cgi.parse_qs(self.rfile.read(length),
                                    keep_blank_values=1)
                args = simplejson.loads(args['data'][0])
            except:
                import traceback
                traceback.print_exc()
                self.wfile.write("Error decoding POST data")
                return traceback.format_exc()
        else:
            args = {}

        if not urldata.path in self.handlers:
            self.wfile.write(self._usage(args))
            return

        args['engine'] = self.engine

        self.wfile.write(self.handlers[urldata.path](args))

        return


class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    pass


class HTTPMonitor(object):
    def __init__(self, stats, harness, port):
        self.port = int(port)
        self.stats = stats
        self.harness = harness
        self.logmanager = harness.logmanager
        self.httpd = None
        self.stopped = False
        self.run_thread = None
        self.new_handlers = {}

    def add_handler(self, path, callback):
        self.new_handlers[path] = callback

    def get_logs(self):
        """
        Pull a list of logfiles for all of the tasks
        that this tasksitter has created.
        """
        return self.logmanager.get_logfile_names()

    def get_stats(self):
        """
        Return metadata and running stats for the process
        """
        data = self.stats.get_metadata()
        data.update(self.stats.get_live_data())
        return data

    def start(self):
        """
        Begin serving HTTP requests with stats data
        """
        if self.stopped:
            return

        self.run_thread = threading.Thread(target=self._start_server,
                                           name="HTTPServer")
        self.run_thread.start()

    def stop(self):
        """
        Stop the HTTP server
        """
        if self.httpd:
            self.httpd.shutdown()

        self.stopped = True

    def _start_server(self):
        """
        Internal method to start the server.
        """
        handler = lambda x, y, z: HTTPMonitorHandler(
            self,
            self.new_handlers, x, y, z)
        self.httpd = ThreadedTCPServer(('', self.port), handler)
        self.httpd.timeout = 1
        self.httpd.serve_forever(poll_interval=0.1)



class MachineProvider(object):
    def get_all_shared_fate_zones(self):
        pass

    def list_available_machines(self, sf_zone):
        pass

    def usable(self):
        return False

    def decomission(self, machine):
        pass



"""
A class to encapsulate data about a process
"""
import datetime
import os
import resource
import signal
import sys
import time


class Process(object):
    """
    An object representing the child process or running task
    """
    def __init__(self, pid):
        self.pid = pid

        self.previous_update_time = 0
        self.last_usage_update = 0
        self.usage = None
        self.last_usage = None
        self.system_usage = None
        self.last_system_usage = None
        self.cpu_usage = 0
        self.mem_usage = None
        self.proc_stats = []
        self.start_time = datetime.datetime.now()

    def is_alive(self):
        """
        Check if the child process is alive
        """
        try:
            os.waitpid(self.pid, os.WNOHANG)
            return True
        except OSError:
            return False
        except KeyboardInterrupt:
            print "WAA" * 100
            self.force_exit()

    def force_exit(self):
        """
        Send a SIGKILL to the child process
        """
        print "Killing Process %s" % self.pid
        try:
            os.kill(self.pid, signal.SIGKILL)
        except OSError:
            # Already dead
            pass

        try:
            os.killpg(self.pid, signal.SIGKILL)
        except OSError:
            # Already dead
            pass

    def wait_for_completion(self):
        """
        Wait for the child process to finish.
        Returns:
         None, None if the child is already dead
         return value of os.waitpid() if we wait for it to exit
        """
        if not self.is_alive():
            return None, None

        try:
            return os.waitpid(self.pid, 0)
        except OSError:
            # child already exited
            return 0, 0

    @classmethod
    def get_system_cpu_usage(cls):
        """
        Read out an array of info from /proc/stat
        utime, nicetime, stime, idle, iowate, irq, softiq
        """

        cpu_stats = file("/proc/stat", "r").readline()
        columns = cpu_stats.replace("cpu", "").split(" ")
        return [int(a) for a in columns if a]

    def get_proc_cpu_usage(self, deep=False):
        """
        Read out an array of info from /proc/pid/stat
        utime, nicetime, stime, idle, iowate, irq, softiq
        """

        def get_proc_cpu(pid, pgrp=None):
            columns = self.proc_stats[pid]
            if not pgrp or pgrp == int(columns[4]):
                return [int(c) for c in
                        [columns[13], 0, columns[14], 0, 0, 0, 0]]
            else:
                return [0] * 7

        if not deep:
            return get_proc_cpu(self.pid)

        # Deep means get the cpu usage for all processes in our pgrp
        # Best way I can figure to do this is to check ALL procs in
        # the kernel process table, sadface.
        cpu_usage = [0] * 7
        for proc in self.proc_stats.keys():
            usage = get_proc_cpu(int(proc), self.pid)
            for index, i in enumerate(usage):
                cpu_usage[index] += i

        return cpu_usage

    def get_proc_stats(self):
        """
        Cache /proc/ID/stat so we only have to read it once
        """
        stats = {}
        for proc in os.listdir('/proc'):
            try:
                raw_stats = file("/proc/%d/stat" % int(proc), "r").readline()
                stats[int(proc)] = raw_stats.split(" ")
            except ValueError:
                # proc isn't a pid
                pass

        return stats

    def get_proc_mem_usage(self, deep=False):
        """
        Get memory usage of the child process
        """

        def get_proc_mem(pid, pgrp=None):
            columns = self.proc_stats[pid]
            if not pgrp or pgrp == int(columns[4]):
                return [int(columns[22]), (int(columns[23])
                                           * resource.getpagesize())]
            else:
                return [0, 0]

        if not deep:
            return get_proc_mem(self.pid)

        # Deep means get the memory for all processes in our pgrp
        # Best way I can figure to do this is to check ALL procs in
        # the kernel process table, sadface.
        mem_usage = [0, 0]
        for proc in self.proc_stats.keys():
            usage = get_proc_mem(proc, self.pid)
            mem_usage[0] += usage[0]
            mem_usage[1] += usage[1]

        # Returns vmem, res
        return mem_usage

    def calculate_cpu_usage(self):
        """
        Calculate CPU Usage for this process.

        Reference: http://stackoverflow.com/questions/1420426/
                   calculating-cpu-usage-of-a-process-in-linux

        Formula:
        utime_jiffys_proc_used / utime_jiffys_system_used +
        stime_jiffys_proc_used / stime_jiffys_system_used

        Return: True if child is in violation, false otherwise
        """
        if not self.usage or not self.last_usage:
            return

        proc_user_time_diff = (self.usage[0] -
                               self.last_usage[0])

        system_user_time_diff = ((self.system_usage[0] +
                                  self.system_usage[1]) -
                                 (self.last_system_usage[0] +
                                  self.last_system_usage[1])) or 1

        proc_sys_time_diff = (self.usage[2] -
                              self.last_usage[2])

        # Include system time and idle time
        system_sys_time_diff = (sum(self.system_usage[2:3]) -
                                sum(self.last_system_usage[2:3])) or 1

        #print "system:", system_sys_time_diff, system_user_time_diff
        #print "proc:", proc_sys_time_diff, proc_user_time_diff

        user_time_perc = float(proc_user_time_diff) / system_user_time_diff
        sys_time_perc = float(proc_sys_time_diff) / system_sys_time_diff

        self.cpu_usage = user_time_perc + sys_time_perc

        #print self.cpu_usage
        return self.cpu_usage

    def update_usage(self, deep=False):
        """Update process usage.

        Only updates at most once per 0.1 seconds

        Return: True if we updated, otherwise False
        """

        now = time.time()
        if now - self.last_usage_update > 0.1:
            self.previous_update_time = self.last_usage_update
            self.last_usage = self.usage

            try:
                self.proc_stats = self.get_proc_stats()
                self.usage = self.get_proc_cpu_usage(deep)
                self.last_system_usage = self.system_usage
                self.system_usage = Process.get_system_cpu_usage()
                self.mem_usage = self.get_proc_mem_usage(deep)
            except IOError:
                # Process died and /proc/PID no longer exists
                return

            self.calculate_cpu_usage()
            self.last_usage_update = time.time()
            return True

        return False



import json
import requests
import sys


def get_help_string():
    return "Update clustersitter logging level"


def get_command():
    return "update_log_level"


def get_parser(parser):
    parser.add_argument("--level", dest="level",
                        required=True,
                        help='New Debug Level '
                        '(10=debug,20=info,30=warning,40=error)',
                        type=int)

    return parser


def run_command(clustersitter_url, level):
    data = {'data': json.dumps({'level': level})}
    print "%s/update_idle_limit" % clustersitter_url
    resp = requests.post("%s/update_logging_level" % clustersitter_url,
                         data=data)
    print resp.content



import json
import requests


def get_help_string():
    return "Update clustersitter job config"


def get_command():
    return "update_job_cfg"


def get_parser(parser):
    parser.add_argument("--job-config", dest="job_config",
                        required=True, nargs='+',
                        help='JSON file of Job configuration')

    return parser


def add_config_from_file(clustersitter_url, filename):
    try:
        jobs = json.load(open(filename))
    except:
        print "Error opening config file %s" % filename
        import traceback
        traceback.print_exc()
        return

    print "%s/add_job" % clustersitter_url
    for job in jobs:
        data = {'data': json.dumps(job)}
        resp = requests.post("%s/add_job" % clustersitter_url, data=data)
        print "Response for %s: %s" % (job['task_configuration']['name'], resp.content)


def run_command(clustersitter_url=None,
                job_config=None):
    if isinstance(job_config, basestring):
        add_config_from_file(clustersitter_url, job_config)
    else:
        for filename in job_config:
            add_config_from_file(clustersitter_url, filename)



import logging
import threading
import time
from datetime import datetime, timedelta

from eventmanager import ClusterEventManager
from jobfiller import JobFiller

logger = logging.getLogger(__name__)


class ProductionJob(object):
    def __init__(self,
                 dns_basename,
                 task_configuration,
                 deployment_layout,
                 deployment_recipe,
                 recipe_options=None,
                 persistent=False,
                 linked_job=None,
                 ):
        """
        Args:
          task_configuration: JSON representing options to pass to a tasksitter
          to launch the job
          deployment_layout: How many CPUS/Memory per shared fate zone,  dict
          deployment_recipe: A python class which knows how to deploy this task
          recipe_options: any options to pass to the recipe
          linked_job: Specify the name of another job.  This job will
          be placed on the same machine.
        """
        # The config to pass to a machinesitter / tasksitter
        self.dns_basename = dns_basename or ""
        self.task_configuration = task_configuration
        self.linked_job = linked_job
        self.linked_job_object = None
        self.name = task_configuration['name']
        self.deployment_recipe = deployment_recipe
        self.recipe_options = recipe_options or {}
        self.sitter = None
        self.currently_spawning = {}
        self.persistent = persistent

        # A mapping of SharedFateZoneName: {'cpu': #CPU, 'mem': MB_Mem_Per_CPU}
        self.deployment_layout = deployment_layout

        self.fillers = {}
        for zone in self.deployment_layout.keys():
            self.fillers[zone] = []
            self.currently_spawning[zone] = 0

        #!MACHINEASSUMPTION!
        # Hack to make num_machines == num_cpu, for now.
        for zone in self.deployment_layout.keys():
            self.deployment_layout[zone]['num_machines'] = \
                self.deployment_layout[zone]['cpu']

    def get_shared_fate_zones(self):
        return self.deployment_layout.keys()

    def get_num_required_machines_in_zone(self, zone, state):
        """
        Return the total number of machines needed in this zone
        """
        #!MACHINEASSUMPTION!
        if not self.linked_job:
            return self.deployment_layout.get(zone, {}).get('num_machines', 0)

        self.find_linked_job(state)

        return self.linked_job_object.deployment_layout.get(
            zone, {}).get('num_machines', 0)

    def get_name(self):
        return self.task_configuration['name']

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self.to_dict())

    def to_dict(self):
        return {
            'task_configuration': self.task_configuration,
            'deployment_layout': self.deployment_layout,
            'deployment_recipe': self.deployment_recipe,
            'recipe_options': self.recipe_options,
            'persistent': self.persistent,
            'linked_job': self.linked_job,
            }

    def do_update_deployment(self, state, version=None):
        """
        1. Find all machines running this job
        2. Build a JobFiller with all those machines
        3. Run the JobFiller, making sure the version is being
           passed to the deployment recipe
        """
        self.recipe_options['version'] = version

        for zone, machines in state.machines_by_zone.items():
            job_machines = []
            for machine in machines:
                tasks = machine.get_running_tasks()
                names = [task['name'] for task in tasks]
                if self.name in names:
                    job_machines.append(machine)

            if job_machines:
                filler = JobFiller(
                    len(job_machines),
                    self,
                    zone,
                    job_machines,
                    reboot_task=True
                    )

                logger.info(
                    "Starting job filler for code update for %s" % self.name)
                filler.start()
                if not zone in self.fillers:
                    self.fillers[zone] = []

                self.fillers[zone].append(filler)

    def get_zone_overflow(self, state):
        zone_overflow = {}

        for zone in self.get_shared_fate_zones():
            zone_overflow[zone] = 0
            required = self.get_num_required_machines_in_zone(zone, state)
            active = state.job_fill.get(self.name, {}).get(zone, 0)

            if active > required:
                zone_overflow[zone] += (active - required)

        return zone_overflow

    def find_dependent_jobs(self):
        dependent_jobs = []
        for job in self.sitter.state.jobs:
            if job.linked_job == self.name:
                dependent_jobs.append(job)

        return dependent_jobs

    def find_linked_job(self, state):
        if self.linked_job_object:
            return self.linked_job_object

        linked_job = None
        for job in state.jobs:
            if job.name == self.linked_job:
                linked_job = job
                break

        self.linked_job_object = linked_job
        return self.linked_job_object


    def ensure_on_linked_job(self, state, sitter):
        """
        1. Ensure the linked job exists, if not bail out
        2. Ensure that this job is running on each machine
        that the linked job is on.  If not, create a job filler for
        those machines and this job.
        Note: As a linked job we should never create a job filler
        that spawns new machines.  We should always just be populating
        existing machines.
        """
        linked_job = self.find_linked_job(state)

        if not linked_job:
            logger.warn("Couldn't find linked job (%s) for %s!" % (
                    self.linked_job, str(self)))
            # Returning False stops all other jobs this cycle, which
            # we don't want to do.
            return True

        for zone in linked_job.get_shared_fate_zones():
            machines_to_fill = []
            machines = state.job_fill_machines[linked_job.name][zone]

            for machine in machines:
                task_names = [
                    task['name'] for task in machine.get_running_tasks()]

                if not self.name in task_names:
                    machines_to_fill.append(machine)

            current_fillers = self.fillers[zone]
            currently_spawning = 0
            for filler in current_fillers:
                currently_spawning += filler.num_remaining()

            # Also check the linked job for active job fillers
            # we don't want to start a filler here if the linked job
            # is also actively filling, it should be sequential.
            current_fillers = linked_job.fillers[zone]
            for filler in current_fillers:
                currently_spawning += filler.num_remaining()

            if not currently_spawning and len(machines_to_fill) > 0:
                ClusterEventManager.handle(
                    "New JobFiller for Linked Job: %s, %s, %s, %s" % (
                        machines_to_fill, zone, str(self), self.linked_job))

                filler = JobFiller(len(machines_to_fill), self,
                                   zone, machines_to_fill)
                filler.start()
                self.fillers[zone].append(filler)

        return True

    def refill(self, state, sitter):
        self.sitter = sitter

        new_machines = False
        while self.sitter.machines_in_queue():
            new_machines = True
            # We want to ensure any machines recently added to monitoring
            # have had a chance to load their data, incase they are
            # running this job
            logger.info("Waiting for machine monitors to load machine data "
                        "before filling jobs")
            time.sleep(0.5)

        if new_machines:
            # If we had to wait for new machines that means that
            # there are new machines, and we need to recalculate
            # job fill before it is safe to do refill.  The next
            # pass should be OK.
            logger.info("Waiting for next jobfill to be calculated before "
                        "doing a refill")

            return False

        while not self.name in state.job_fill:
            # 1) Assume this job has already been added to state.jobs
            # 2) Want to ensure calculator has run at least once to find out
            #    if this job already exists throughout the cluster
            logger.info("Waiting for calculator thread to kick in before "
                         "filling jobs")
            time.sleep(0.5)

        # Clear out finished fillers after 5 minutes
        for zone, fillers in self.fillers.items():
            for filler in fillers:
                now = datetime.now()
                if (filler.is_done() and
                    now - filler.end_time > timedelta(minutes=5)):
                    logger.info("Removing a filler from %s for %s" % (
                            zone, self.name))
                    self.fillers[zone].remove(filler)

        # If we have a linked job then bypass all the normal logic
        # and just piggyback on those machines
        if self.linked_job:
            return self.ensure_on_linked_job(state, sitter)

        #!MACHINEASSUMPTION!
        # Step 1: Ensure we have enough machines in each SFZ
        # Step 1a: Check for idle machines and reserve as we find them
        for zone in self.get_shared_fate_zones():
            idle_available = state.get_idle_machines_in_zone(zone)
            total_required = self.get_num_required_machines_in_zone(zone, state)
            idle_required = total_required - state.job_fill[self.name][zone]

            current_fillers = self.fillers[zone]
            currently_spawning = 0
            for filler in current_fillers:
                currently_spawning += filler.num_remaining()

            self.currently_spawning[zone] = currently_spawning

            idle_required -= currently_spawning

            # !MACHINEASSUMPTION! Ideally we're counting resources here
            # not machines
            required_new_machine_count = max(
                (idle_required - len(idle_available)), 0)

            do_log = logger.debug
            if idle_required > 0:
                do_log = logger.info

            do_log(
                ("Calculated job requirements for %s in %s: " % (self.name,
                                                                 zone)) +
                "Currently Active: %s " % (state.job_fill[self.name][zone]) +
                "Idle Required: %s, Total New: %s " % (
                    idle_required,
                    required_new_machine_count) +
                "Currently Spawning: %s " % (currently_spawning) +
                "idle-available: %s " % (len(idle_available)) +
                "total_required: %s " % (total_required)
                )

            usable_machines = []
            if required_new_machine_count <= 0:
                # idle_available > idle_required, so use just as many
                # as we need
                usable_machines = idle_available[:idle_required]
            elif required_new_machine_count > 0:
                # Otherwise take all the available idle ones, and
                # we'll make more
                usable_machines.extend(idle_available)

            if idle_required > 0:
                ClusterEventManager.handle(
                    "New JobFiller: %s, %s, %s, %s" % (
                        idle_required, zone, str(self), usable_machines))

                filler = JobFiller(idle_required, self,
                                   zone, usable_machines)
                self.fillers[zone].append(filler)
                filler.start()

        return True



#!/usr/bin/python
"""
This is the main file of the tasksitter application.

The task sitter's job is to spawn a job and the monitor
it's resource usage.  If it exceeds the specified parameters it reboots
the job.

If a job is 'flapping' or being rebooted too often it knows how to trigger
an alert.
"""
import os
import simplejson
import sys


import sittercommon.arg_parser as argparse
import sittercommon.http_monitor as http_monitor
import sittercommon.logmanager as logmanager
import constraints
import process_harness
import stats_collector


def run_command_with_harness(command, args, constraints_list):
    """Execute the child command.

    Args:
      command: a string which includes the file and args
      args: An object which containts configuration options
      constraints: an array of Constraint objects

    Return:
      A Harness object encapsulating the child process
      and all the current constraints
    """

    logs = logmanager.LogManager(args.stdout_location,
                                 args.stderr_location)

    return process_harness.ProcessHarness(command, constraints_list,
                                          restart=args.restart,
                                          max_restarts=args.max_restarts,
                                          poll_interval=args.poll_interval,
                                          collect_stats=args.collect_stats,
                                          logmanager=logs,
                                          uid=args.uid)


def parse_args(args):
    """ Parse command line args."""

    parser = argparse.ArgumentParser(description="Run a task with a"
                                     "cpu/memory harness")
    parser.add_argument('--cpu', dest='cpu', type=float,
                        help='The amount of cores (in float) '
                        'this task can use')

    parser.add_argument('--mem', dest='mem', type=float,
                        help='The amount of memory in MB that this '
                        'task can use')

    parser.add_argument('--time-limit', dest='time_limit', type=float,
                        help='Maximum time the child can run for in seconds')

    parser.add_argument('--disable_stat_collection', dest='collect_stats',
                        action='store_false',
                        default=True,
                        help='Disable collecting child process stats unless'
                        'required by a constraint.  Saves some CPU overhead.')

    parser.add_argument('--restart', dest='restart',
                        action='store_true',
                        default=False,
                        help='Restart the task if it violates any of its '
                        'constraints')

    parser.add_argument('--max-restarts', dest='max_restarts',
                        default=-1, type=int,
                        help='Number of times to reboot the task when it '
                        'violates constraints before bailing out.')

    parser.add_argument('--ensure-alive', dest='ensure_alive',
                        default=False,
                        action='store_true',
                        help='Restart the task if it exists normally.  A '
                        'normal exit does incremement the restart counter')

    parser.add_argument('--http-monitoring', dest='http_monitoring',
                        default=False,
                        action='store_true',
                        help='Expose an interface via HTTP for collecting '
                        'task statistics and metadata')

    parser.add_argument('--http-monitoring-port', dest='http_monitoring_port',
                        default=8080, type=int,
                        help='Port to do HTTP Monitoring (Default: 80)')

    parser.add_argument('--keep-http-running', dest='keep_http_running',
                        default=False,
                        action='store_true',
                        help='Keep http server up after children die')

    parser.add_argument('--command', dest='command',
                        required=True,
                        help='The command to run')

    parser.add_argument('--poll-interval', dest='poll_interval',
                        default=0.1, type=float,
                        help='How frequently (seconds) to poll the child '
                        'process for constraint violations '
                        '(default=0.1 seconds)')

    parser.add_argument('--stdout-location', dest='stdout_location',
                        default='-', type=str,
                        help='Directory where stdout logs should be placed '
                        'default is to print to caller\'s STDOUT')

    parser.add_argument('--stderr-location', dest='stderr_location',
                        default='-', type=str,
                        help='Directory where stdout logs should be placed '
                        'default is to print to caller\'s STDERR')

    parser.add_argument('--uid', dest='uid',
                        help='Change to UID before executing child process'
                        'requires root priviledges.  Can be an ID or name.')

    # explicitly offer args the param incase we're parsing not from
    # sys.argv
    return parser.parse_args(args=args)


def build_constraints(args):
    """
    Build an array of Constraint objects based on invokation ars.
    """
    proc_constraints = []

    if args.ensure_alive:
        proc_constraints.append(constraints.LivingConstraint())

    if args.cpu:
        proc_constraints.append(constraints.CPUConstraint(args.cpu))

    if args.mem:
        proc_constraints.append(constraints.MemoryConstraint(args.mem))

    if args.time_limit:
        proc_constraints.append(constraints.TimeConstraint(args.time_limit))

    return proc_constraints


def main(sys_args=None, wait_for_child=True, allow_spam=False):
    """Run the task sitter."""

    if not sys_args:
        sys_args = sys.argv[1:]

    print sys_args

    args = parse_args(sys_args)
    constraints_list = build_constraints(args)

    # Set outselves to our own pgrp to separate from machine sitter
    os.setpgrp()

    # Ensure any previous instances of this task have died
    procs = [(int(p), c) for p, c in [x.rstrip('\n').split(' ', 1) \
               for x in os.popen('ps h -eo pid:1,command')]]
    for proc in procs:
      if args.command in proc[1]:
        if proc[0] != os.getpid():
          print "Killing prexisting task: %s:%s" % proc
          try:
              os.kill(proc[0], 9)
          except:
              import traceback
              traceback.print_exc()

    harness = run_command_with_harness(args.command, args, constraints_list)
    harness.allow_spam = allow_spam
    harness.begin_monitoring()

    stats = stats_collector.StatsCollector(harness)
    httpd = None

    if args.http_monitoring:
        httpd = http_monitor.HTTPMonitor(stats, harness,
                                         args.http_monitoring_port)
        httpd.start()

    if wait_for_child:
        exit_code = harness.wait_for_child_to_finish()

        print simplejson.dumps(harness.logmanager.get_logfile_names())

        if httpd and not args.keep_http_running:
            httpd.stop()

        if not args.keep_http_running:
            sys.exit(exit_code)

    else:
        return stats, httpd, harness

if __name__ == '__main__':
    main()
